From b2cc0e805c4b41ece8042432276e456d1e958956 Mon Sep 17 00:00:00 2001
From: James Hunter <hunjmes@amazon.com>
Date: Tue, 8 Sep 2020 15:59:19 +0000
Subject: [PATCH] Fix initialization of parallel BTree scan

Before, function _bt_first() would exit immediately if the specified scan
keys could never be satisfied--without notifying other parallel workers, if
any, that the scan key was done. This moved that particular worker to a
scan key beyond what was in the shared parallel-query state, so that it
would later try to read in "InvalidBlockNumber", without recognizing it as
a special sentinel value.

The basic bug is that the BTree parallel query state machine assumes that a
worker process is working on a key <= the global key--a worker process can
be behind (i.e., hasn't finished its work on a previous key), but never
ahead. By allowing the first worker to move on to the next scan key, in
this one case, without notifying other workers, the global key ends up <
the first worker's local key.

---
 src/backend/access/nbtree/nbtsearch.c | 4 ++++
 1 file changed, 4 insertions(+)

diff --git a/src/backend/access/nbtree/nbtsearch.c b/src/backend/access/nbtree/nbtsearch.c
index 1e628a33d7..8f6575fdf1 100644
--- a/src/backend/access/nbtree/nbtsearch.c
+++ b/src/backend/access/nbtree/nbtsearch.c
@@ -880,7 +880,11 @@ _bt_first(IndexScanDesc scan, ScanDirection dir)
 	 * never be satisfied (eg, x == 1 AND x > 2).
 	 */
 	if (!so->qual_ok)
+	{
+		/* Notify any other workers that we're done with this scan key. */
+		_bt_parallel_done(scan);
 		return false;
+	}
 
 	/*
 	 * For parallel scans, get the starting page from shared state. If the
-- 
2.15.3.AMZN

